#pragma once

#include <array>
#include <string>
#include <iostream>
#include <type_traits>
#include <list>
#include <vector>

#include <boost/lexical_cast.hpp>
#include <boost/numeric/conversion/cast.hpp>
#include <boost/tokenizer.hpp>

#include <boost/property_tree/ptree.hpp>

#include <dune/common/fvector.hh>

#include <dune/amdis/Output.hpp>
#include <dune/amdis/Math.hpp>

namespace AMDiS
{
  namespace detail
  {
//     double mu_parser_eval(std::string const& valStr);

    template <class T> inline T numeric_cast(double value)
    {
      return boost::numeric_cast< T >(value);
    }

    template <> inline bool numeric_cast<bool>(double value)
    {
      return value != 0.0;
    }

    /// convert string to intrinsic type
    template <class T, class Enable = void>
    struct Convert
    {
      static void eval(std::string valStr, T& value)
      {
        value = boost::lexical_cast<T>(valStr);
      }
    };

//     template <class T>
//     struct Convert<T, std::enable_if_t<std::is_arithmetic<T>::value> >
//     {
//       static void eval(std::string valStr, T& value)
//       {
//         try {
//             value = numeric_cast< T >(mu_parser_eval(valStr));
//         } catch(...) {
//             error_exit("Could not parse '", valStr, "' to '", typeid(T).name(), "'\n");
//         }
//       }
//     };

//     template <class T>
//     struct Convert<T, std::enable_if_t<std::is_enum<T>::value> >
//     {
//       static void eval(std::string valStr, T& value)
//       {
//         EnumParser<T>()(valStr, value);
//       }
//     };

//     convert string to vector
    template <class V, int dim>
    struct Convert<Dune::FieldVector<V, dim>>
    {
      using T = Dune::FieldVector<V, dim>;
      static void eval(std::string valStr, T& values)
      {
        using value_type = typename T::value_type;
        using Tokenizer = boost::tokenizer<boost::char_separator<char>>;

        boost::char_separator<char> sep(",; ");
        Tokenizer tokens(valStr, sep);
        int i = 0;
        for (auto token : tokens)
        {
          test_exit(i < dim, "Vector data exceeds field-vector dimension!");

          value_type v;
          Convert<value_type>::eval(token, v);
          values[i++] = v;
        }
      }
    };

//     convert string to vector
    template <class V, size_t dim>
    struct Convert<std::array<V, dim>>
    {
      using T = std::array<V, dim>;
      static void eval(std::string valStr, T& values)
      {
        using value_type = typename T::value_type;
        using Tokenizer = boost::tokenizer<boost::char_separator<char>>;

        boost::char_separator<char> sep(",; ");
        Tokenizer tokens(valStr, sep);
        size_t i = 0;
        for (auto token : tokens)
        {
          test_exit(i < dim, "Vector data exceeds array dimension!");

          value_type v;
          Convert<value_type>::eval(token, v);
          values[i++] = v;
        }
      }
    };

    // convert string to vector
    template <class T, class Alloc>
    struct Convert<std::vector<T, Alloc>>
    {
      static void eval(std::string valStr, std::vector<T>& values)
      {
        using value_type = T;
        using Tokenizer = boost::tokenizer<boost::char_separator<char>>;

        boost::char_separator<char> sep(",; ");
        Tokenizer tokens(valStr, sep);
        for (auto token : tokens)
        {
          value_type v;
          Convert<value_type>::eval(token, v);
          values.push_back(v);
        }
      }
    };

  } // end namespace detail


  /// output-stream for std::list
  template <class T, class Alloc>
  std::ostream& operator<<(std::ostream& out, std::list<T,Alloc> const& l)
  {
    auto it = l.begin();
    out << "["; if (l.size() > 0) out << *it;
    for (; it != l.end(); ++it)
      out << ", " << *it;
    out << "]";
    return out;
  }


  /// output-stream for std::vector
  template <class T, class Alloc>
  std::ostream& operator<<(std::ostream& out, std::vector<T,Alloc> const& l)
  {
    auto it = l.begin();
    out << "["; if (l.size() > 0) out << *it;
    for (; it != l.end(); ++it)
      out << ", " << *it;
    out << "]";
    return out;
  }

  inline void replaceAll(std::string& str, std::string const& from, std::string const& to)
  {
    if (from.empty())
      return;
    size_t start_pos = 0;
    while ((start_pos = str.find(from, start_pos)) != std::string::npos)
    {
      str.replace(start_pos, from.length(), to);
      start_pos += to.length();
    }
  }

  // _____________________________________________________________________________

  /** Basis data container as a map of tag on a value as strings. The container
  *  throws an exception, if the tag was not found.
  */
  struct Initfile
  {
    using Self = Initfile;

    /** initialize init-file from file with filename in, read data and save it
    *  to singleton-map
    *  @param in: filename string
    */
    static void init(std::string in);


    /** Static get routine for getting parameter-values from init-file
    *  initialized in init()-method.
    *  Cast the value to the desired type using std::stringstream.
    *  @param tag: The tag to look for
    *  @param value: The result.
    *  @param debugInfo: msgInfo for current parameter. (0..no printing,
    *    1..print missing parameter info, 2..print parameter value) [optional]
    */
    template <class T>
    static void get(std::string tag, T& value, int debugInfo = -1)
    {
      if (debugInfo == -1)
        debugInfo = singlett().getMsgInfo();
      else
      {
        int swap(debugInfo);
        debugInfo = singlett().getMsgInfo();
        singlett().msgInfo=swap;
      }

      using path = boost::property_tree::ptree::path_type;
      replaceAll(tag, "->", ">");
      auto tagPath = path(tag, '>');

      // TODO: use boost::optional instead
      // TODO: use convert method from above
      std::string valueStr = "-";
      valueStr = singlett().pt.get(tagPath, valueStr);

      if (valueStr != "-")
        detail::Convert<T>::eval(valueStr, value);

//       if (debugInfo == 2)
//       {
//         std::cout << "Parameter '" << tag << "'"
//                   << " initialized with: " << value << std::endl;
//       }
      singlett().msgInfo = debugInfo;
    }


    template <class T, class S>
    static T get(std::string tag, S const& default_value)
    {
      T value = default_value;
      Self::get(tag, value);
      return value;
    }


    template <class T>
    static T get(std::string tag)
    {
      T value; nullify(value);
      Self::get(tag, value);
      return value;
    }


    /// update map tag->value_old to tag->value in singleton
    template <class T>
    static void set(std::string tag, T& value, int debugInfo=  -1)
    {
      if (debugInfo == -1)
        debugInfo = singlett().getMsgInfo();

      using path = boost::property_tree::ptree::path_type;
      replaceAll(tag, "->", ">");
      auto tagPath = path(tag, '>');
      singlett().pt.put(tagPath, value);

      // update msg parameters msgInfo, msgWait, paramInfo
      singlett().getInternalParameters();
//       if (debugInfo == 2)
//         std::cout << "Parameter '" << tag << "'"
//                   << " set to: " << value << std::endl;
    }


    /// add map tag->value to data in singleton
    template <class T>
    static void add(std::string tag, T& value, int debugInfo = -1)
    {
      set(tag, value, debugInfo);
    }

    /// Returns specified info level
    static int getMsgInfo()
    {
      return singlett().msgInfo;
    }


    /// print all data in singleton to std::cout
    static void printParameters();


    /// clear data in singleton
    static void clearData()
    {
      // singlett().clear();
    }


    /// save singlett-data to file with filename fn
    static void save(std::string /*fn*/)
    {
      using namespace boost::property_tree;
      // json_parser::write_jason(fn, singlett().pt);
    }

  protected:
    Initfile() = default;

    /// return the singleton that contains the data
    static Initfile& singlett()
    {
      static Initfile initfile;
      return initfile;
    }

    /** Fill the initfile from an input stream.
    * @param in: the stream to fill the data from.
    * Current dataformat: tag:value
    * Comment char: percent '%'
    * Include files: #include "filename" or #include <filename>
    */
    void read(std::string fn, bool force = false);

    /// Write data-map to initfile with filename fn
    void write(std::string fn);

    /// read parameters for msgInfo, msgWait, paramInfo
    void getInternalParameters();

    int msgInfo = 0;
    int paramInfo = 1;
    int breakOnMissingTag = 0;

    /// boost:property_tree to read/store parameter values
    boost::property_tree::ptree pt;
  };

  using Parameters = Initfile;

#ifndef AMDIS_NO_EXTERN_INITFILE
  extern template void Initfile::get(std::string, int&, int);
  extern template void Initfile::get(std::string, double&, int);
  extern template void Initfile::get(std::string, std::string&, int);
#endif

} // end namespace AMDiS
