#pragma once

#include <list>
#include <vector>

#include <dune/geometry/type.hh>

#include <dune/amdis/OperatorTermBase.hpp>
#include <dune/amdis/terms/TermGenerator.hpp>

#include <dune/amdis/common/TypeDefs.hpp>
#include <dune/amdis/utility/GetDegree.hpp>

namespace AMDiS
{
  enum FirstOrderType
  {
    GRD_PSI,
    GRD_PHI
  };


  template <class MeshView>
  class Operator
  {
    using Self = Operator;
    using OperatorTermType = OperatorTerm<MeshView>;

    using ElementVector = Impl::ElementVector;
    using ElementMatrix = Impl::ElementMatrix;

    using IdType = typename MeshView::Grid::LocalIdSet::IdType;

  public:
    /// Add coefficients for zero-order operator < c * u, v >.
    /// The coefficient \p c must be a scalar expression.
    template <class Term>
    Self& addZOT(Term const& c)
    {
      return addZOTImpl(toTerm(c));
    }

    template <class... Args>
    static shared_ptr<Self> zot(Args&&... args)
    {
      auto op = make_shared<Self>();
      op->addZOT(std::forward<Args>(args)...);
      return op;
    }


    /// Add coefficients for first-order operator < psi, term * grad(phi) > or
    /// < grad(psi), term * phi >, where the first corresponds to
    /// \p firstOrderType == GRD_PHI and the second one to \p firstOrderType == GRD_PSI.
    /// The coefficient \p b must be a vector expression.
    template <class Term>
    Self& addFOT(Term const& b, FirstOrderType firstOrderType)
    {
      return addFOTImpl(toTerm(b), firstOrderType);
    }

    /// Add coefficients for first-order operator < psi, b * d_i(phi) > or
    /// < d_i(psi), b * phi >, where the first corresponds to
    /// \p firstOrderType == GRD_PHI and the second one to \p firstOrderType == GRD_PSI.
    /// The coefficient \p b must be a scalar expression.
    template <class Term>
    Self& addFOT(Term const& b, std::size_t i, FirstOrderType firstOrderType)
    {
      return addFOTImpl(toTerm(b), i, firstOrderType);
    }

    template <class... Args>
    static shared_ptr<Self> fot(Args&&... args)
    {
      auto op = make_shared<Self>();
      op->addFOT(std::forward<Args>(args)...);
      return op;
    }


    /// Add coefficients for second-order operator < grad(psi), A * grad(phi) >,
    /// where \p A can be a matrix or a scalar expression.
    template <class Term>
    Self& addSOT(Term const& A)
    {
      return addSOTImpl(toTerm(A));
    }

    /// Add coefficients for second-order operator < d_i(psi), A * d_j(phi) >,
    /// where \p A can be a matrix or a scalar expression. The first index \p _i
    /// corresponds to the derivative component of the test-function and the
    /// second index \p _j to the derivative component of the trial function.
    template <class Term>
    Self& addSOT(Term const& A, std::size_t i, std::size_t j)
    {
      return addSOTImpl(toTerm(A), i, j);
    }

    template <class... Args>
    static shared_ptr<Self> sot(Args&&... args)
    {
      auto op = make_shared<Self>();
      op->addSOT(std::forward<Args>(args)...);
      return op;
    }


    /// Calls \ref zot(), \ref for() or \ref sot(), depending on template
    /// parameter \p Order.
    template <size_t Order, class... Args>
    static shared_ptr<Self> create(Args&&... args)
    {
      return create(index_<Order>{}, std::forward<Args>(args)...);
    }


    /// Extract the polynomial degree from \p rowFeSpace and \p colFeSpace.
    template <class RowFeSpace, class ColFeSpace>
    void init(RowFeSpace const& rowFeSpace,
              ColFeSpace const& colFeSpace);


    /// Calculates the needed quadrature degree for the given term-order \p order.
    /// This takes into account the polynomial degree of trial and test functions
    /// and the polynomial degree of the coefficient functions
    template <class Element>
    int getQuadratureDegree(Element const& element, int order, FirstOrderType firstOrderType = GRD_PHI);


    template <class RowView, class ColView>
    bool getElementMatrix(RowView const& rowView,
                          ColView const& colView,
                          ElementMatrix& elementMatrix,
                          double* factor = NULL);


    template <class RowView>
    bool getElementVector(RowView const& rowView,
                          ElementVector& elementVector,
                          double* factor = NULL);

  protected:
    template <class RowView, class ColView>
    void assembleZeroOrderTerms(RowView const& rowView,
				ColView const& colView,
				ElementMatrix& elementMatrix);

    template <class RowView>
    void assembleZeroOrderTerms(RowView const& rowView,
				ElementVector& elementVector);

    template <class RowView, class ColView>
    void assembleFirstOrderTermsGrdPhi(RowView const& rowView,
				       ColView const& colView,
				       ElementMatrix& elementMatrix);

    template <class RowView, class ColView>
    void assembleFirstOrderTermsGrdPsi(RowView const& rowView,
				       ColView const& colView,
				       ElementMatrix& elementMatrix);

    template <class RowView>
    void assembleFirstOrderTermsGrdPsi(RowView const& rowView,
                                       ElementVector& elementVector);

    template <class RowView, class ColView>
    void assembleSecondOrderTerms(RowView const& rowView,
				  ColView const& colView,
				  ElementMatrix& elementMatrix);

  private:
    template <class Term>
    Self& addZOTImpl(Term const& term);

    template <class Term>
    Self& addFOTImpl(Term const& term, FirstOrderType firstOrderType);

    template <class Term>
    Self& addFOTImpl(Term const& term, std::size_t i, FirstOrderType firstOrderType);

    template <class Term>
    Self& addSOTImpl(Term const& term);

    template <class Term>
    Self& addSOTImpl(Term const& term, std::size_t i, std::size_t j);


    template <class... Args>
    static shared_ptr<Self> create(index_<0>, Args&&... args)
    {
      return zot(std::forward<Args>(args)...);
    }

    template <class... Args>
    static shared_ptr<Self> create(index_<1>, Args&&... args)
    {
      return fot(std::forward<Args>(args)...);
    }

    template <class... Args>
    static shared_ptr<Self> create(index_<2>, Args&&... args)
    {
      return sot(std::forward<Args>(args)...);
    }

  private:
    template <class LocalView>
    IdType getElementId(LocalView const& localView);

  private:
    /// List of all second order terms
    std::list<OperatorTermType*> secondOrder;

    /// List of all first order terms derived to psi
    std::list<OperatorTermType*> firstOrderGrdPsi;

    /// List of all first order terms derived to phi
    std::list<OperatorTermType*> firstOrderGrdPhi;

    /// List of all zero order terms
    std::list<OperatorTermType*> zeroOrder;

    int psiDegree = 1;
    int phiDegree = 1;

    IdType lastMatrixId = 0;
    IdType lastVectorId = 0;

    ElementMatrix cachedElementMatrix;
    ElementVector cachedElementVector;
  };

} // end namespace AMDiS

#include "Operator.inc.hpp"
