#pragma once

#include <cassert>

#include <dune/amdis/common/ValueCategory.hpp>

namespace AMDiS {


template <class K, int SIZE>
K sum(Dune::FieldVector<K, SIZE> const& vector)
{
  K result = 0;
  for (int i = 0; i < SIZE; ++i)
    result += vector[i];
  return result;
}

template <class T, int N>
Dune::FieldMatrix<T, N, N> outer(Dune::FieldVector<T,N> const& vec1, Dune::FieldVector<T,N> const& vec2)
{
  Dune::FieldMatrix<T, N, N> mat;
  for (int i = 0; i < N; ++i)
    for (int j = 0; j < N; ++j)
      mat[i][j] = vec1[i] * vec2[j];
  return mat;
}


template <class T, int N>
T inner(Dune::FieldVector<T,N> const& vec1, Dune::FieldVector<T,N> const& vec2)
{
  T result = 0;
  for (int i = 0; i < N; ++i)
    result += vec1[i] * vec2[i];
  return result;
}


template <class T, int N>
Dune::FieldVector<T,N> operator*(Dune::FieldVector<T,N> vec1, T scalar)
{
  return vec1 *= scalar;
}

template <class T, int N>
Dune::FieldVector<T,N> operator*(T scalar, Dune::FieldVector<T,N> vec1)
{
  return vec1 *= scalar;
}


template <class T, int N, int M>
Dune::FieldVector<T,N> operator*(Dune::FieldMatrix<T,N,M> const& mat, Dune::FieldVector<T,M> const& vec)
{
  return Dune::FMatrixHelp::mult(mat, vec);
}


class OperatorEvaluation
{
  using Scalar = Dune::FieldVector<double, 1>;

protected:
  template <class... Args>
  auto evalZotImpl(Args...) const { assert( false ); return 0; }

  template <class ValueType>
  auto evalZotImpl(tag::scalar, tag::none, ValueType const& v,
                    Scalar const& test, Scalar const& trial) const
  {
    return v * test * trial;
  }

  template <class... Args>
  auto evalFotImpl(Args...) const { assert( false ); return 0; }

  template <class ValueType, class Gradient>
  auto evalFotImpl(tag::scalar, VectorComponent comp, ValueType const& v,
                    Gradient const& grad_test, Scalar const& trial) const
  {
    return v * grad_test[comp.index] * trial;
  }

  template <class ValueType, class Gradient>
  auto evalFotImpl(tag::vector, tag::none, ValueType const& v,
                    Gradient const& grad_test, Scalar const& trial) const
  {
    return (v * grad_test) * trial;
  }

  template <class ValueType, class Gradient>
  auto evalFotImpl(tag::scalar, tag::none, ValueType const& v,
                    Gradient const& grad_test, Scalar const& trial) const
  {
    return v * sum(grad_test) * trial;
  }


  template <class... Args>
  auto evalSotImpl(Args...) const { assert( false ); return 0; }

  template <class ValueType, class Gradient>
  auto evalSotImpl(tag::scalar, MatrixComponent comp, ValueType const& v,
                    Gradient const& grad_test, Gradient const& grad_trial) const
  {
    return v * (grad_test[comp.row] * grad_trial[comp.col]);
  }

  template <class ValueType, class Gradient>
  auto evalSotImpl(tag::matrix, tag::none, ValueType const& M,
                    Gradient const& grad_test, Gradient const& grad_trial) const
  {
    return grad_test * (M * grad_trial);
  }

  template <class ValueType, class Gradient>
  auto evalSotImpl(tag::scalar, tag::none, ValueType const& v,
                    Gradient const& grad_test, Gradient const& grad_trial) const
  {
    return v * (grad_test * grad_trial);
  }
};


} // end namespace AMDiS
