#pragma once

#include <dune/amdis/ProblemInstatBase.hpp>
#include <dune/amdis/ProblemStat.hpp>
#include <dune/amdis/common/Utility.hpp>

namespace AMDiS
{
  // forward declarations
  class AdaptInfo;

  /**
   * \ingroup Problem
   *
   * \brief
   * Standard implementation of ProblemTimeInterface for a time
   * dependent problems.
   */
  template <class Traits>
  class ProblemInstat : public ProblemInstatBase
  {
    using Self        = ProblemInstat;
    using ProblemType = ProblemStat<Traits>;

    using SystemVectorType = typename ProblemType::SystemVectorType;

  public:
    /// Constructs a ProblemInstat with prob as its stationary problem.
    ProblemInstat(std::string name, ProblemType& prob)
      : ProblemInstatBase(name, nullptr)
      , problemStat(prob)
    {}

    ProblemInstat(std::string name, ProblemType& prob, ProblemStatBase& initialProb)
      : ProblemInstatBase(name, &initialProb)
      , problemStat(prob)
    {}

    /// Initialisation of the problem.
    virtual void initialize(Flag initFlag = INIT_NOTHING);

    /// Used in \ref initialize().
    virtual void createUhOld();

    /// Implementation of \ref ProblemTimeInterface::initTimestep().
    virtual void initTimestep(AdaptInfo& adaptInfo) override;

    /// Implementation of \ref ProblemTimeInterface::closeTimestep().
    virtual void closeTimestep(AdaptInfo& adaptInfo) override;

    /// Returns \ref problemStat.
    ProblemType& getStatProblem()
    {
      return problemStat;
    }

    /// Returns \ref oldSolution.
    decltype(auto) getOldSolution()       { return *oldSolution; }
    decltype(auto) getOldSolution() const { return *oldSolution; }

    /// Returns the I'th component of \ref oldSolution.
    template <size_t I = 0>
    decltype(auto) getOldSolution(const index_<I> _i = {})
    {
        return (*oldSolution)[_i];
    }

    /// Implementation of \ref ProblemTimeInterface::transferInitialSolution().
    virtual void transferInitialSolution(AdaptInfo& adaptInfo) override;

  protected:
    /// Space problem solved in each timestep.
    ProblemType& problemStat;

    /// Solution of the last timestep.
    std::shared_ptr<SystemVectorType> oldSolution;
  };


#ifndef AMDIS_NO_EXTERN_PROBLEMINSTAT
  extern template class ProblemInstat<ProblemStatTraits<2>>;
#endif

} // end namespace AMDiS

#include "ProblemInstat.inc.hpp"
