/** \file concepts_base.hpp */

#pragma once

// std c++ headers
#include <utility>
#include <type_traits>

// AMDiS headers
#include <dune/amdis/common/Mpl.hpp>
#include <dune/amdis/common/Utility.hpp>

// macro to generate concept-checks
#define HAS_MEMBER_GENERATE(name)                                           \
  template <class F, class... Args>                                         \
  inline auto invoke_member_ ## name(F&& f, Args&&... args) ->              \
    decltype(std::forward<F>(f).name (std::forward<Args>(args)...)) {       \
    return std::forward<F>(f).name (std::forward<Args>(args)...);           \
  }                                                                         \
  inline no_valid_type invoke_member_ ## name(...);                         \
  template <class, class> struct has_member_ ## name : std::false_type {};           \
  template <class F, class Return, class... Args>                           \
  struct has_member_ ## name <F, Return(Args...)>                           \
    : std::is_convertible<                                                  \
        decltype(invoke_member_ ## name(std::declval<F>(), std::declval<Args>()...)), \
        Return > {}; 

#define HAS_MEMBER(name) has_member_ ## name

#include <dune/functions/common/functionconcepts.hh>
        
namespace AMDiS
{
  namespace traits
  {
    // Some type traits to test for type-attributes of a class
    // -------------------------------------------------------------------------

    namespace Impl
    {
      template <class T, class = typename T::value_type >
      std::true_type HasValueTypeImpl(int);

      template <class T> std::false_type HasValueTypeImpl(...);


      template <class T, class = typename T::size_type >
      std::true_type HasSizeTypeImpl(int);

      template <class T> std::false_type HasSizeTypeImpl(...);


      template <class T, class = typename T::result_type >
      std::true_type HasResultTypeImpl(int);

      template <class T> std::false_type HasResultTypeImpl(...);

    } // end namespace Impl


    template <class T>
    using HasValueType = decltype( Impl::HasValueTypeImpl<T>(int{}) );

    template <class T>
    using HasSizeType = decltype( Impl::HasSizeTypeImpl<T>(int{}) );

    template <class T>
    using HasResultType = decltype( Impl::HasResultTypeImpl<T>(int{}) );

    
    // -------------------------------------------------------------------------
             
                              
    namespace Impl 
    {
      template <class F, class Arg>                                         
      inline auto invoke_vector(F&& f, Arg&& arg)
        -> decltype( std::forward<F>(f)[std::forward<Arg>(arg)] );

      inline no_valid_type invoke_vector(...);
        
    } // end namespace Impl
    
    template <class, class>
    struct IsVector : std::false_type {};

    template <class F, class Return, class Arg>
    struct IsVector<F, Return(Arg)>
        : std::is_same< decltype( Impl::invoke_vector(std::declval<std::decay_t<F>>(), std::declval<Arg>()) ),
                        Return > {};

    template <class, class>
    struct IsVectorWeak : std::false_type {};

    template <class F, class Return, class Arg>
    struct IsVectorWeak<F, Return(Arg)>
        : std::is_convertible< decltype( Impl::invoke_vector(std::declval<std::decay_t<F>>(), std::declval<Arg>()) ),
                               Return > {}; 
                               
                               
    namespace Impl
    {
      template <class T, class = decltype(&T::push_back())>
      std::true_type PushBackImpl(int);

      template <class T> std::false_type PushBackImpl(...);

    } // end namespace Impl
    
    template <class T> 
    using HasPushBack = decltype(Impl::PushBackImpl<T>(int{}));
    
  } // end namespace traits


  /// Namespace that implements basic concept checks
  namespace Concepts
  {
    /** 
      * \defgroup Concepts Concepts definitions
      * @{
      **/

    /// A Collable is a function `F` that can be called with arguments of type `Args...`.
    /**
      * To be used as follows: `Concepts::Collable<F, Args...>`. Returns true, if
      * an instance of `F` can be called by `operator()` with arguments of type
      * `Args...`.
      **/
    template <class F, class... Args>
    constexpr bool Callable = Dune::Functions::Concept::isCallable<F, Args...>();
    
    
    /// A Functor is a function `F` with signature `Signature`.
    /**
      * To be used as follows: `Concepts::Functor<F, R(Args...)>`. Returns true, if
      * an instance of `F` can be called by `operator()` with arguments of type
      * `Args...` and returns a value of type `R`, i.e. `Signature := R(Args...)`.
      **/
    template <class F, class Signature> // F, Signature=Return(Arg)
    constexpr bool Functor = Dune::Functions::Concept::isFunction<F, Signature>();
    
    
    /// A predicate is a function that returns a boolean.
    template <class F, class... Args>
    constexpr bool Predicate = Functor<F, bool(Args...)>;
    
    
    /// Test whether a type `T` supports vector-access `vector[size_t]`.
    /**
      * To be used as follows: `Concepts::Vector<V>`. Returns true, if `V` has a 
      * value type, accessible by `Value_t<V>` and an instance of `V` can be called
      * by square brackets `operator[]` with an index of type `size_t`. The return
      * value of this element access should be convertible to `Value_t<V>`.
      **/
    template <class T>
    constexpr bool Vector = traits::IsVectorWeak<T, Value_t<T>(size_t)>::value;

    
    /// Test whether a type `T` supports matrix-access `matrix(size_t, size_t)`.
    /**
      * To be used as follows: `Concepts::Matrix<M>`. Returns true, if `M` has a 
      * value type, accessible by `Value_t<M>` and an instance of `M` can be called
      * by round brackets `operator()` with two indices of type `size_t`. The return
      * value of this element access should be convertible to `Value_t<V>`.
      **/
    template <class T>
    constexpr bool Matrix = Functor<T, Value_t<T>(size_t, size_t)>;
    
    /** @} **/

  } // end namespace Concepts

} // end namespace AMDiS
