#pragma once

#include <dune/istl/solvers.hh>
#include <dune/istl/umfpack.hh>

namespace AMDiS
{
  // creators for solvers, depending on matrix-type
  // ---------------------------------------------------------------------------

  template <class Matrix, class VectorX, class VectorB = VectorX>
  class ISTLSolverCreator;

  template <class... MatrixRows, class VectorX, class VectorB>
  class ISTLSolverCreator<Dune::MultiTypeBlockMatrix<MatrixRows...>, VectorX, VectorB>
  {
    using BaseSolver = Dune::InverseOperator<VectorX, VectorB>;

  public:
    ISTLSolverCreator(std::string prefix)
      : prefix(prefix)
      , info(2)
    {
      Parameters::get(prefix + "->info", info);
    }

    template <class Matrix, class Precon>
    unique_ptr<BaseSolver> create(std::string name, Matrix& A, Precon& P)
    {
      size_t maxIter = 500;
      Parameters::get(prefix + "->max iteration", maxIter);

      double rTol = 1.e-6;
      Parameters::get(prefix + "->reduction", rTol);

      if (name == "cg")
        return make_unique<Dune::CGSolver<VectorX>>(A, P, rTol, maxIter, info);
      else if (name == "bicgstab")
        return make_unique<Dune::BiCGSTABSolver<VectorX>>(A, P, rTol, maxIter, info);
      else if (name == "minres")
        return make_unique<Dune::MINRESSolver<VectorX>>(A, P, rTol, maxIter, info);
      else if (name == "gmres")
      {
        int restart = 30;
        Parameters::get(prefix + "->restart", restart);
        return make_unique<Dune::RestartedGMResSolver<VectorX>>(A, P, rTol, restart, maxIter, info);
      }
      else
        error_exit("Unknown solver name!");
    }

  private:
    std::string prefix;
    int         info;
  };


  template <class Block, class Alloc, class VectorX, class VectorB>
  class ISTLSolverCreator<Dune::BCRSMatrix<Block,Alloc>, VectorX, VectorB>
  {
    using BaseSolver = Dune::InverseOperator<VectorX, VectorB>;
    using Matrix     = Dune::BCRSMatrix<Block,Alloc>;

  public:
    ISTLSolverCreator(std::string prefix)
      : prefix(prefix)
      , info(2)
    {
      Parameters::get(prefix + "->info", info);
    }

    template <class LinOperator, class Precon>
    unique_ptr<BaseSolver> create(std::string name, LinOperator& A, Precon& P)
    {
      size_t maxIter = 500;
      Parameters::get(prefix + "->max iteration", maxIter);

      double rTol = 1.e-6;
      Parameters::get(prefix + "->relative tolerance", rTol);

      if (name == "cg")
        return make_unique<Dune::CGSolver<VectorX>>(A, P, rTol, maxIter, info);
      else if (name == "bicgstab")
        return make_unique<Dune::BiCGSTABSolver<VectorX>>(A, P, rTol, maxIter, info);
      else if (name == "minres")
        return make_unique<Dune::MINRESSolver<VectorX>>(A, P, rTol, maxIter, info);
      else if (name == "gmres")
      {
        int restart = 30;
        Parameters::get(prefix + "->restart", restart);
        return make_unique<Dune::RestartedGMResSolver<VectorX>>(A, P, rTol, restart, maxIter, info);
      }
#if HAVE_SUITESPARSE_UMFPACK
      else if (name == "umfpack" || name == "direct")
        return make_unique<Dune::UMFPack<Matrix>>(A, info);
#endif
      else
        error_exit("Unknown solver name!");
    }

  private:
    std::string prefix;
    int         info;
  };

} // end namespace AMDiS
