#pragma once


#include <dune/amdis/linear_algebra/LinearSolverInterface.hpp>
#include <dune/amdis/linear_algebra/istl/ISTLRunner.hpp>

namespace AMDiS
{
  // A general solver class for istl solvers
  // ---------------------------------------------------------------------------

  template <class Matrix, class VectorX, class VectorB, class Runner>
  class LinearSolver
    : public LinearSolverInterface<Matrix, VectorX, VectorB>
  {
    using Super = LinearSolverInterface<Matrix, Vector, Vector>;
    using RunnerBase = typename Super::RunnerBase;

  public:
    /// Constructor
    LinearSolver(std::string name, std::string prefix)
      : runner(make_shared<Runner>(name, prefix))
    {}


  private:
    /// Implements \ref LinearSolverInterface::solveImpl()
    virtual void solveImpl(Matrix const& A, VectorX& x, VectorB const& b,
                           SolverInfo& solverInfo) override
    {
      Timer t;
      if (solverInfo.doCreateMatrixData()) {
        // init matrix or wrap block-matrix or ...
        runner->init(A);
      }

      // solve the linear system
      int error = runner->solve(A, x, b, solverInfo);
      solverInfo.setError(error);

      if (!solverInfo.doStoreMatrixData())
        runner->exit();
    }

  private:
    /// redirect the implementation to a runner. Implements a \ref RunnerInterface
    shared_ptr<Runner> runner;
  };


  template <class Matrix, class VectorX, class VectorB>
  class SolverCreator<Matrix, VectorX, VectorB>
  {
    using SolverBase = LinearSolverInterface<Matrix, VectorX, VectorB>;

    template <class Runner>
    using Solver = LinearSolver<Matrix, VectorX, VectorB, Runner>;

    /// Instantiate a new linear solver
    static unique_ptr<SolverBase> create(std::string name, std::string prefix)
    {
      using ISTL = ISTLRunner<Matrix, VectorX, VectorB>;

      if (name == "cg" || name == "bicgstab" || name == "minres" || name == "gmres")
        return make_unique<Solver<ISTL>>(name, prefix);
      else
        error_exit("Unknown solver name!"); // TODO: add direct solver interface here
    }
  };


} // end namespace AMDiS
