#pragma once

#include <vector>
#include <string>
#include <memory>
#include <tuple>

#include <dune/istl/multitypeblockmatrix.hh>
#include <dune/istl/multitypeblockvector.hh>

#include <dune/amdis/Output.hpp>
#include <dune/amdis/common/TupleUtility.hpp>
#include <dune/amdis/common/Utility.hpp>
#include <dune/amdis/linear_algebra/istl/DOFMatrix.hpp>

// for explicit instantiation
#include <dune/amdis/ProblemStatTraits.hpp>

namespace AMDiS
{
  namespace Impl
  {
    // DOFMatrices = std::tuple< std::tuple< DOFMatrix<RowFeSpace, ColFeSpace, ValueType>... >... >
    template <class DOFMatrices>
    struct BuildDOFMatrices
    {
      template <size_t R>
      using DOFMatrixRow = std::tuple_element_t<R, DOFMatrices>;

      template <size_t R, size_t C>
      using DOFMatrixRowCol = std::tuple_element_t<C, DOFMatrixRow<R>>;

      // add arg to repeated constructor argument list
      template <size_t... R, class RowFeSpaces, class ColFeSpaces, class MultiMatrix>
      static DOFMatrices make(Indices<R...>,
                              RowFeSpaces&& rowFeSpaces,
                              ColFeSpaces&& colFeSpace,
                              MultiMatrix&& multiMatrix)
      {
        return DOFMatrices{make_row(
          index_<R>(),
          MakeSeq_t<std::tuple_size<std::decay_t<ColFeSpaces>>::value>(),
          std::get<R>(std::forward<RowFeSpaces>(rowFeSpaces)),
          std::forward<ColFeSpaces>(colFeSpace),
          std::get<R>(std::forward<MultiMatrix>(multiMatrix))
          )...};
      }

      template <size_t R, size_t... C, class RowFeSpace, class ColFeSpaces, class MultiVector>
      static DOFMatrixRow<R> make_row(index_<R>,
                                      Indices<C...>,
                                      RowFeSpace&& rowFeSpace,
                                      ColFeSpaces&& colFeSpace,
                                      MultiVector&& multiVector)
      {
        return DOFMatrixRow<R>{DOFMatrixRowCol<R,C>(
          std::forward<RowFeSpace>(rowFeSpace),
          std::get<C>(std::forward<ColFeSpaces>(colFeSpace)),
          "matrix[" + std::to_string(R) + "][" + std::to_string(C) + "]",
          std::get<C>(std::forward<MultiVector>(multiVector))
          )...};
      }
    };

    template <class DOFMatrices, class RowFeSpaces, class ColFeSpaces, class MultiMatrix>
    DOFMatrices buildDOFMatrices(RowFeSpaces&& rowFeSpaces,
                                  ColFeSpaces&& colFeSpace,
                                  MultiMatrix&& multiMatrix)
    {
        return BuildDOFMatrices<DOFMatrices>::make(
            MakeSeq_t<std::tuple_size<std::decay_t<RowFeSpaces>>::value>(), // traverse rows first
            std::forward<RowFeSpaces>(rowFeSpaces),
            std::forward<ColFeSpaces>(colFeSpace),
            std::forward<MultiMatrix>(multiMatrix));
    }

  } // end namespace Impl




  /// \brief Container that repesents multiple data-Vectors of different value types.
  /**
    *  Represents a \ref Dune::MultiTypeBlockVector + a tuple of corresponding
    *  feSpaces. This I'th \ref Dune::BlockVector compbined with the I'th FeSpace
    *  builds a \ref DOFVector  and can be return by the \ref operator[].
    *
    *  By default, the ValueTypes are all \ref Dune::FieldMatrix of 1x1 double entry.
    **/
  template <class FeSpaces,
            class FieldType  = double,
            class Dimensions = Repeat_t<std::tuple_size<FeSpaces>::value, index_<1>> >
  class SystemMatrix
  {
    template <class RowFeSpace, class RowDim>
    struct MultiMatrixRowGenerator
    {
      template <class ColDim>
      using ValueTypeGenerator = Dune::FieldMatrix<FieldType, RowDim::value, ColDim::value>;
      using ValueTypes = MakeTuple_t<ValueTypeGenerator, Dimensions>;

      template <class ColFeSpace, class VT>
      using DOFMatrixGenerator = DOFMatrix<RowFeSpace, ColFeSpace, VT>;
      using DOFMatrices  = MakeTuple2_t<DOFMatrixGenerator, FeSpaces, ValueTypes>;

      template <class VT>
      using BaseMatrixGenerator = Dune::BCRSMatrix<VT, std::allocator<VT>>;
      using BaseMatrices = MakeTuple_t<BaseMatrixGenerator, ValueTypes>;

      using MultiVector = ExpandTuple_t<Dune::MultiTypeBlockVector, BaseMatrices>;
    };

    template <class RowFeSpace, class RowDim>
    using MultiVectorGenerator = typename MultiMatrixRowGenerator<RowFeSpace, RowDim>::MultiVector;

    template <class RowFeSpace, class RowDim>
    using ValueTypeGenerator = typename MultiMatrixRowGenerator<RowFeSpace, RowDim>::ValueTypes;

    template <class RowFeSpace, class RowDim>
    using DOFMatrixGenerator = typename MultiMatrixRowGenerator<RowFeSpace, RowDim>::DOFMatrices;

  public:
    using MultiVectors = MakeTuple2_t<MultiVectorGenerator, FeSpaces, Dimensions>;
    using ValueTypes   = MakeTuple2_t<ValueTypeGenerator, FeSpaces, Dimensions>;
    using DOFMatrices  = MakeTuple2_t<DOFMatrixGenerator, FeSpaces, Dimensions>;
    using MultiMatrix  = ExpandTuple_t<Dune::MultiTypeBlockMatrix, MultiVectors>;

    static constexpr size_t nComponents = std::tuple_size<FeSpaces>::value;

    AMDIS_STATIC_ASSERT( nComponents > 0 );
    AMDIS_STATIC_ASSERT( nComponents == std::tuple_size<Dimensions>::value );

    /// Constructor that constructs new base-vectors
    SystemMatrix(FeSpaces const& feSpaces)
      : feSpaces(feSpaces)
      , matrix{}
      , dofmatrices(Impl::buildDOFMatrices<DOFMatrices>(feSpaces, feSpaces, matrix))
    {}


    /// Return the number of vector entries
    static constexpr size_t N()
    {
      return std::tuple_size<FeSpaces>::value;
    }

    static constexpr size_t M()
    {
      return std::tuple_size<FeSpaces>::value;
    }

    /// Return a shared pointer to the i'th underlaying base vector
    template <size_t R, size_t C>
    auto& getMatrix(const index_<R> = index_<R>(),
                    const index_<C> = index_<C>())
    {
      static_assert(R < N() && C < M(), "Indices out of range [0,N)x[0,M)");
      return std::get<C>(std::get<R>(matrix));
    }

    /// Return a shared pointer to the i'th underlaying base vector
    template <size_t R, size_t C>
    auto const& getMatrix(const index_<R> = index_<R>(),
                          const index_<C> = index_<C>()) const
    {
      static_assert(R < N() && C < M(), "Indices out of range [0,N)x[0,M)");
      return std::get<C>(std::get<R>(matrix));
    }

    /// Return the underlying multi vector
    MultiMatrix& getMatrix()
    {
        return matrix;
    }

    MultiMatrix const& getMatrix() const
    {
      return matrix;
    }

    /// Return the I'th finite element space
    template <size_t I = 0>
    auto& getRowFeSpace(const index_<I> = index_<I>()) const
    {
      static_assert(I < N(), "Index out of range [0,N)");
      return std::get<I>(feSpaces);
    }

    template <size_t I = 0>
    auto const& getColFeSpace(const index_<I> = index_<I>()) const
    {
      static_assert(I < M(), "Index out of range [0,M)");
      return std::get<I>(feSpaces);
    }

    template <size_t R = 0, size_t C = 0>
    auto& operator()(const index_<R> = index_<R>(),
                      const index_<C> = index_<C>())
    {
      static_assert(R < N() && C < M(), "Indices out of range [0,N)x[0,M)");
      return std::get<C>(std::get<R>(dofmatrices));
    }

    /// Create a DOFVector with references to the unerlaying data
    template <size_t R = 0, size_t C = 0>
    auto const& operator()(const index_<R> = index_<R>(),
                            const index_<C> = index_<C>()) const
    {
      static_assert(R < N() && C < M(), "Indices out of range [0,N)x[0,M)");
      return std::get<C>(std::get<R>(dofmatrices));
    }

  private:
    FeSpaces const& feSpaces;	///< a tuple of feSpaces
    MultiMatrix 	matrix;		///< a tuple of data-vectors
    DOFMatrices     dofmatrices;
  };


  // specialization for 1 component
  // -------------------------------------------------------------------------

  template <class FeSpace,
            class FieldType,
            int dim>
  class SystemMatrix<std::tuple<FeSpace>, FieldType, std::tuple<index_<dim>>>
  {
    using ValueType     = Dune::FieldMatrix<FieldType, dim, dim>;
    using DOFMatrixType = DOFMatrix<FeSpace, FeSpace, ValueType>;

  public:
    using MultiMatrix   = typename DOFMatrixType::BaseMatrix;

    static constexpr size_t nComponent = 1;


    /// Constructor that constructs new base-vectors
    template <class FeSpaces>
    SystemMatrix(FeSpaces const& feSpaces)
      : dofmatrix(std::get<0>(feSpaces), std::get<0>(feSpaces), "")
    {}


    /// Return the number of vector entries
    static constexpr size_t N()
    {
      return 1;
    }

    static constexpr size_t M()
    {
      return 1;
    }

    /// Return a shared pointer to the i'th underlaying base vector
    template <size_t R = 0, size_t C = 0>
    MultiMatrix& getMatrix(const index_<R> = index_<R>(),
                            const index_<C> = index_<C>())
    {
      static_assert(R == 0 && C == 0, "Indices out of range [0,1)x[0,1)");
      return dofmatrix.getMatrix();
    }

    /// Return a shared pointer to the i'th underlaying base vector
    template <size_t R = 0, size_t C = 0>
    MultiMatrix const& getMatrix(const index_<R> = index_<R>(),
                                  const index_<C> = index_<C>()) const
    {
      static_assert(R == 0 && C == 0, "Indices out of range [0,1)x[0,1)");
      return dofmatrix.getMatrix();
    }

    /// Return the underlying multi vector
    MultiMatrix const& getMatrix() const
    {
      return dofmatrix.getMatrix();
    }

    MultiMatrix& getMatrix()
    {
      return dofmatrix.getMatrix();
    }

    /// Return the I'th finite element space
    template <size_t I = 0>
    FeSpace const& getRowFeSpace(const index_<I> = index_<I>()) const
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofmatrix.getRowFeSpace();
    }

    template <size_t I = 0>
    FeSpace const& getColFeSpace(const index_<I> = index_<I>()) const
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofmatrix.getColFeSpace();
    }

    template <size_t R = 0, size_t C = 0>
    DOFMatrixType& operator()(const index_<R> = index_<R>(),
                              const index_<C> = index_<C>())
    {
      static_assert(R == 0 && C == 0, "Indices out of range [0,1)x[0,1)");
      return dofmatrix;
    }

    /// Create a DOFVector with references to the unerlaying data
    template <size_t R = 0, size_t C = 0>
    DOFMatrixType const& operator()(const index_<R> = index_<R>(),
                                    const index_<C> = index_<C>()) const
    {
      static_assert(R == 0 && C == 0, "Indices out of range [0,1)x[0,1)");
      return dofmatrix;
    }

  private:
    DOFMatrixType  dofmatrix;
  };

#ifndef AMDIS_NO_EXTERN_SYSTEMMATRIX
    // explicit instantiation in SystemMatrix.cpp
    extern template class SystemMatrix<typename ProblemStatTraits<2,2,1>::FeSpaces>;
    extern template class SystemMatrix<typename ProblemStatTraits<2,2,1,2>::FeSpaces>;
#endif

} // end namespace AMDiS
