#pragma once

#include <vector>
#include <string>
#include <memory>
#include <tuple>

#include <dune/istl/multitypeblockvector.hh>

#include <dune/amdis/Output.hpp>
#include <dune/amdis/common/Loops.hpp>
#include <dune/amdis/common/TupleUtility.hpp>
#include <dune/amdis/common/Utility.hpp>

#include <dune/amdis/linear_algebra/istl/DOFVector.hpp>

// for explicit instantiation
#include <dune/amdis/ProblemStatTraits.hpp>

namespace AMDiS
{

  namespace Impl
  {
    // DOFVectors = std::tuple< DOFVector<FeSpace, ValueType>... >
    template <class DOFVectors>
    struct BuildDOFVectors
    {
      template <size_t I>
      using DOFVectorIdx = std::tuple_element_t<I, DOFVectors>;

      template <size_t... I, class FeSpaces, class MultiVector>
      static DOFVectors make(Indices<I...>,
                              FeSpaces&& feSpaces,
                              std::vector<std::string> const& names,
                              MultiVector&& multiVector)
      {
        return DOFVectors{DOFVectorIdx<I>(
          std::get<I>(std::forward<FeSpaces>(feSpaces)),
          names[I],
          std::get<I>(std::forward<MultiVector>(multiVector))
          )...};
      }
    };

    template <class DOFVectors, class FeSpaces, class MultiVector>
    DOFVectors buildDOFVectors(FeSpaces&& feSpaces,
                                    std::vector<std::string> const& names,
                                    MultiVector&& multiVector)
    {
        return BuildDOFVectors<DOFVectors>::make(
            MakeSeq_t<std::tuple_size<std::decay_t<FeSpaces>>::value>(),
            std::forward<FeSpaces>(feSpaces),
            names,
            std::forward<MultiVector>(multiVector));
    }

  } // end namespace Impl


  /// \brief Container that repesents multiple data-Vectors of different value types.
  /**
    *  Represents a \ref Dune::MultiTypeBlockVector + a tuple of corresponding
    *  feSpaces. This I'th \ref Dune::BlockVector compbined with the I'th FeSpace
    *  builds a \ref DOFVector  and can be return by the \ref operator[].
    *
    *  By default, the ValueTypes are all \ref Dune::FieldVector of 1 double entry.
    **/
  template <class FeSpaces,
            class ValueTypes = Repeat_t<std::tuple_size<FeSpaces>::value, Dune::FieldVector<double,1>> >
  class SystemVector
  {
    using Self = SystemVector;

  public:
    template <class T> using BlockVectorGenerator = Dune::BlockVector<T, std::allocator<T>>;

    using BaseVectors = MakeTuple_t<BlockVectorGenerator, ValueTypes>;
    using MultiVector = ExpandTuple_t<Dune::MultiTypeBlockVector, BaseVectors>;

    using DOFVectors  = MakeTuple2_t<DOFVector, FeSpaces, ValueTypes>;

    static constexpr size_t nComponents = std::tuple_size<FeSpaces>::value;

    AMDIS_STATIC_ASSERT( nComponents > 0 );
    AMDIS_STATIC_ASSERT( nComponents == std::tuple_size<ValueTypes>::value );

    /// Constructor that constructs new base-vectors
    SystemVector(FeSpaces const& feSpaces, std::vector<std::string> const& names)
      : feSpaces(feSpaces)
      , names(names)
      , vector{}
      , dofvectors(Impl::buildDOFVectors<DOFVectors>(feSpaces, names, vector))
    {
      compress();
    }


    /// Return the number of vector entries
    static constexpr size_t size()
    {
      return std::tuple_size<FeSpaces>::value;
    }

    /// Return number of elements
    int count() const
    {
      return int(size());
    }

    /// Return a shared pointer to the i'th underlaying base vector
    template <size_t I>
    auto& getVector(const index_<I> = index_<I>())
    {
      static_assert(I < size(), "Index out of range [0,SIZE)" );
      return std::get<I>(vector);
    }

    /// Return a shared pointer to the i'th underlaying base vector
    template <size_t I>
    auto const& getVector(const index_<I> = index_<I>()) const
    {
      static_assert(I < size(), "Index out of range [0,SIZE)" );
      return std::get<I>(vector);
    }

    /// Return the underlying multi vector
    MultiVector& getVector()
    {
      return vector;
    }

    MultiVector const& getVector() const
    {
      return vector;
    }

    /// Return the I'th finite element space
    template <size_t I = 0>
    auto const& getFeSpace(const index_<I> = index_<I>()) const
    {
      static_assert(I < size(), "Index out of range [0,SIZE)" );
      return std::get<I>(feSpaces);
    }

    /// Return the name of the i'th DOFVector
    std::string getName(size_t i) const
    {
        return names[i];
    }

    template <size_t I>
    auto& operator[](const index_<I>)
    {
      static_assert(I < size(), "Index out of range [0,SIZE)" );
      return std::get<I>(dofvectors);
    }

    template <size_t I>
    auto const& operator[](const index_<I>) const
    {
      static_assert(I < size(), "Index out of range [0,SIZE)" );
      return std::get<I>(dofvectors);
    }

    template <size_t I>
    auto& getDOFVector(const index_<I> = index_<I>())
    {
      static_assert(I < size(), "Index out of range [0,SIZE)" );
      return std::get<I>(dofvectors);
    }

    template <size_t I>
    auto const& getDOFVector(const index_<I> = index_<I>()) const
    {
      static_assert(I < size(), "Index out of range [0,SIZE)" );
      return std::get<I>(dofvectors);
    }

    /// Resize the I'th \ref vector to the sizes of the I'th \ref feSpaces.
    template <size_t I = 0>
    void compress(const index_<I> = index_<I>())
    {
      std::get<I>(dofvectors).compress();
    }

    /// Resize the \ref vectors to the sizes of the \ref feSpaces.
    void compress()
    {
      For<0, size()>::loop([this](const auto I) {
        std::get<I>(dofvectors).compress();
      });
    }

    void copy(Self const& that)
    {
      For<0, size()>::loop([this, &that](const auto I) {
        std::get<I>(dofvectors).copy(that[I]);
      });
    }

    template <class Scalar>
    std::enable_if_t< std::is_arithmetic<Scalar>::value, Self&>
    operator*=(Scalar s)
    {
      For<0, size()>::loop([this, s](const auto I) {
        std::get<I>(vector) *= s;
      });
      return *this;
    }

    template <class Scalar>
    std::enable_if_t< std::is_arithmetic<Scalar>::value, Self&>
    operator=(Scalar s)
    {
      For<0, size()>::loop([this, s](const auto I) {
        std::get<I>(vector) = s;
      });
      return *this;
    }

  private:
    FeSpaces const& feSpaces;	///< a tuple of feSpaces
    std::vector<std::string> names;

    MultiVector vector;		///< a tuple of data-vectors
    DOFVectors	dofvectors;	///< a tuple of dofvectors referencing \ref vector
  };


  // specialization for the case of only 1 component.

  template <class FeSpace, class ValueType>
  class SystemVector<std::tuple<FeSpace>, std::tuple<ValueType>>
  {
    using Self = SystemVector;

  public:
    using MultiVector = Dune::BlockVector<ValueType>;
    using DOFVectors  = DOFVector<FeSpace, ValueType>;

    static constexpr size_t nComponent = 1;

    /// Constructor that constructs new base-vectors
    template <class FeSpaces>
    SystemVector(FeSpaces const& feSpaces, std::vector<std::string> const& names)
      : dofvector(std::get<0>(feSpaces), names.front())
    {
      compress();
    }


    /// Return the number of vector entries
    static constexpr size_t size()
    {
      return 1;
    }

    /// Return number of elements
    int count() const
    {
      return 1;
    }

    /// Return a shared pointer to the i'th underlaying base vector
    template <size_t I = 0>
    MultiVector& getVector(const index_<I> = index_<I>())
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofvector.getVector();
    }

    /// Return a shared pointer to the i'th underlaying base vector
    template <size_t I = 0>
    MultiVector const& getVector(const index_<I> = index_<I>()) const
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofvector.getVector();
    }

    /// Return the underlying multi vector
    MultiVector& getVector()
    {
      return dofvector.getVector();
    }

    MultiVector const& getVector() const
    {
      return dofvector.getVector();
    }

    /// Return the I'th finite element space
    template <size_t I = 0>
    auto const& getFeSpace(const index_<I> = index_<I>()) const
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofvector.getFeSpace();
    }

    /// Return the name of the i'th DOFVector
    std::string getName(size_t i = 0) const
    {
      test_exit(i == 0, "Index out of range [0,1)");
      return dofvector.getName();
    }

    template <size_t I>
    auto& operator[](const index_<I>)
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofvector;
    }

    template <size_t I>
    auto const& operator[](const index_<I>) const
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofvector;
    }

    template <size_t I>
    auto& getDOFVector(const index_<I> = index_<I>())
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofvector;
    }

    template <size_t I>
    auto const& getDOFVector(const index_<I> = index_<I>()) const
    {
      static_assert(I == 0, "Index out of range [0,1)");
      return dofvector;
    }

    /// Resize the I'th \ref vector to the sizes of the I'th \ref feSpaces.
    template <size_t I = 0>
    void compress(const index_<I> = index_<I>())
    {
      static_assert(I == 0, "Index out of range [0,1)");
      dofvector.compress();
    }

    /// Resize the \ref vectors to the sizes of the \ref feSpaces.
    void compress()
    {
      dofvector.compress();
    }

    void copy(Self const& that)
    {
      dofvector.copy(that[index_<0>()]);
    }

    template <class Scalar>
    std::enable_if_t< std::is_arithmetic<Scalar>::value, Self&>
    operator*=(Scalar s)
    {
      dofvector.getVector() *= s;
      return *this;
    }

    template <class Scalar>
    std::enable_if_t< std::is_arithmetic<Scalar>::value, Self&>
    operator=(Scalar s)
    {
      dofvector.getVector() = s;
      return *this;
    }

  private:
    DOFVectors      dofvector;  ///< a tuple of sofvectors referencing \ref vector
  };




#ifndef AMDIS_NO_EXTERN_SYSTEMVECTOR
    // explicit instantiation in SystemVector.cpp
    extern template class SystemVector<typename ProblemStatTraits<2,2,1>::FeSpaces>;
    extern template class SystemVector<typename ProblemStatTraits<2,2,1,2>::FeSpaces>;
#endif

} // end namespace AMDiS
