#pragma once

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <vector>
#include <type_traits>

#include <dune/istl/bvector.hh>

#include <dune/amdis/OperatorTermBase.hpp>
#include <dune/amdis/common/IndexSeq.hpp>
#include <dune/amdis/common/Traits.hpp>

namespace AMDiS
{
  /**
    * \addtogroup OperatorTerms
    * @{
    **/


  /// \brief An expression that evaluates to the current coordinate of a dof or
  /// quadrature point with given index. \see eval.
  template <class Functor, class... Terms>
  class FunctorTerm
  {
  public:
    using value_type = typename std::result_of<Functor(Value_t<Terms>...)>::type;

    std::tuple<Terms...> terms;

    template <class F, class... Ts,
      class = std::enable_if_t<Concepts::Compatible<Functor, F>> >
    FunctorTerm(F&& f, Ts&&... ts)
      : fct(std::forward<F>(f))
      , terms(std::forward<Ts>(ts)...)
    {}

    /// \brief Evaluates the functor \ref fct at all quadrature points and
    /// stores the result in a vector \ref values.
    template <class Element, class PointList>
    void init(Element const& element,
              PointList const& points)
    {
      AMDIS_FUNCNAME("FunctorTerm::init()");
      for_<0, sizeof...(Terms)>([&,this](auto const _i) {
        std::get<_i>(this->terms).init(element, points);
      });
    }

    value_type operator[](size_t iq) const
    {
      return eval(iq, MakeSeq_t<sizeof...(Terms)>{});
    }

    int getDegree(Dune::GeometryType const& /*t*/) const
    {
      return 2;
    }

  private:
    template <size_t... Is>
    value_type eval(size_t iq, Indices<Is...>) const
    {
      return fct(std::get<Is>(terms)[iq]...);
    }

  private:
    Functor fct;
  };


  /// Generator function for \ref FunctorTerm expressions
  template <class F, class... Ts>
  FunctorTerm< std::decay_t<F>, std::decay_t<Ts>... > func(F&& f, Ts&&... terms)
  {
    return {std::forward<F>(f), std::forward<Ts>(terms)...};
  }

  /** @} **/

} // end namespace AMDiS
