/** \file TermGenerator.hpp */

#pragma once

// std c++ includes
#include <type_traits>

// AMDiS includes
#include <dune/amdis/terms/ConstantTerm.hpp>
#include <dune/amdis/terms/CoordsTerm.hpp>

#include <dune/amdis/common/ConceptsBase.hpp>
#include <dune/amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  namespace Impl
  {
    // annonymous namespace to hide tags
    namespace
    {
      struct _constant {};
      struct _functor {};
      struct _default {};
    }


    template <class T>
    constexpr bool ConstantCategory =
      !std::is_same<ValueCategory_t<T>, tag::unknown>::value;


    // TODO: generalize to functors with arbitrary return values
#ifndef AMDIS_DOW
    template <class T>
    constexpr bool FunctorCategory =
      Concepts::Functor<T, double(Dune::FieldVector<double, 1>)> ||
      Concepts::Functor<T, double(Dune::FieldVector<double, 2>)> ||
      Concepts::Functor<T, double(Dune::FieldVector<double, 3>)>;
#else
    template <class T>
    constexpr bool FunctorCategory =
      Concepts::Functor<T, double(Dune::FieldVector<double, AMDIS_DOW>)>;
#endif


    // TODO: add concept for terms
    template <class T>
    using TermCategory =
      std::conditional_t< ConstantCategory<T>, _constant,
      std::conditional_t< FunctorCategory<T>,  _functor,
                          /* OperatorTerms */  _default > >;


    // forward declaration
    template <class T, class Tag> struct ToTerm;

    template <class T>
    struct ToTerm<T, _constant>
    {
      using type = ConstantTerm<T>;

      static type get(T const& t)
      {
        return {t};
      }
    };

    // if the argument is a functor, that takes a WorldVector
    // and returns a double, see concept \ref CoordsFunctor.
    template <class F>
    struct ToTerm<F, _functor>
    {
      using type = CoordsFunctorTerm<F>;

      template <class F_>
      static type get(F_&& f)
      {
        return {std::forward<F_>(f)};
      }
    };

    // if the argument is already a term return it directly.
    template <class T>
    struct ToTerm<T, _default>
    {
      using type = T const&;

      static constexpr type get(T const& t)
      {
        return t;
      }
    };

  } // end namespace Impl


  /// Generator class that takes a type as argument and returns the corresponding term-type.
  /** The arguments are transformed into terms in the following way:
    * - If type `T` is arithmetic or any small vector/matrix type, generate a
    *   constant term, see \ref ConstantTerm
    * - If the type `T` models a Functor concepts, that maps a WorldVector to a
    *   double type, construct a \ref CoordsTerm
    * - In all other cases, the argument is retured without modification.
    **/
  template <class T>
  using ToTerm = Impl::ToTerm<std::decay_t<T>, Impl::TermCategory<std::decay_t<T>> >;

  template <class T>
  using ToTerm_t = std::decay_t<typename ToTerm<T>::type>;


  /// Generator function for term transformation, \see ToTerm
  template <class T>
  inline typename ToTerm<T>::type toTerm(T&& t)
  {
    return ToTerm<T>::get(std::forward<T>(t));
  }

} // end namespace AMDiS
