#pragma once

#include <nanoflann.hpp>

#include <cstdlib>
#include <iostream>


namespace AMDiS 
{
  template <class FeSpace>
  struct KDTreeMeshAdaptor
  {
    using MeshView = typename FeSpace::GridView;
    using SizeType = typename FeSpace::LocalIndexSet::size_type;
    using ValueType = double;
    
    static constexpr dimension = MeshView::dimension;
    static constexpr dimensionworld = MeshView::dimensionworld;
    
    using Codim0    = typename GridView::template Codim<0>;
    using PointType = typename Codim0::Geometry::GlobalCoordinate;
    using VectorOfPointsType = std::vector<PointType> ;
    
    using DataType = typename MeshView::Grid::EntitySeed;
    using VectorOfDataType = std::vector<DataType>;
    
    using Distance = nanoflann::metric_L2_Simple;
    
    using Self = KDTreeMeshAdaptor<FeSpace>;
    using Metric = typename Distance::template traits<ValueType, Self>::distance_t;
    using IndexType = nanoflann::KDTreeSingleIndexAdaptor<Metric, Self, dimensionworld, SizeType>;

  public:
    unique_ptr<IndexType> index; //! The kd-tree index for the user to call its methods as usual with any other FLANN index.
    FeSpace const& feSpace;

    VectorOfPointsType_ m_points;
    VectorOfDataType_ m_data;
    
    std::map<std::hash<DataType>, DataType> shapeValues;

  public:
    /// Constructor: takes a const ref to the vector of vectors object with the data points
    KDTreeMeshAdaptor(FeSpace const& feSpace_,
		      const int leaf_max_size = 10)
      : feSpace(feSpace_)
      , localView(feSpace.localView())
      , localIndexSet(feSpace.localIndexSet())
    {
      init();
      
      index = make_unique<IndexType>(dimensionworld, *this /* adaptor */, 
        nanoflann::KDTreeSingleIndexAdaptorParams(leaf_max_size, dimensionworld));
      index->buildIndex();
    }

    /// Destructor.
    ~KDTreeMeshAdaptor() {}
    
    void init()
    {
      for (auto const& element : elements(feSpace.gridView())) {
        m_points.push_back(element.geometry().center());
        m_data.push_back(element.seed());
      }
    }
    
    void reinit(const int leaf_max_size = 10)
    {
      m_points.clear();
      m_data.clear();
      init();
      index.reset(new IndexType(dimensionworld, *this /* adaptor */, 
        nanoflann::KDTreeSingleIndexAdaptorParams(leaf_max_size, dimensionworld));
      index->buildIndex();
    }

    /** Query for the \a num_closest closest points to a given point (entered as query_point[0:dim-1]).
      *  Note that this is a short-cut method for index->findNeighbors().
      *  The user can also call index->... methods as desired.
      * \note nChecks_IGNORED is ignored but kept for compatibility with the original FLANN interface.
      */
    inline void query(const ValueType *query_point, 
		      const size_t num_closest, 
		      SizeType *out_indices, 
		      ValueType *out_distances_sq, 
		      const int nChecks_IGNORED = 10) const
    {
      nanoflann::KNNResultSet<ValueType, SizeType> resultSet(num_closest);
      resultSet.init(out_indices, out_distances_sq);
      index->findNeighbors(resultSet, query_point, nanoflann::SearchParams());
    }

    /** @name Interface expected by KDTreeSingleIndexAdaptor
      * @{ */

    Self const& derived() const 
    {
      return *this;
    }
    Self& derived() 
    {
      return *this;
    }

    // Must return the number of data points
    inline size_t kdtree_get_point_count() const 
    {
      return m_points.size();
    }

    // Returns the distance between the vector "p1[0:size-1]" and the data point with index "idx_p2" stored in the class:
    inline ValueType kdtree_distance(const double *p1, 
				      const SizeType idx_p2, 
				      size_t size) const
    {
      ValueType s = 0.0;
      for (size_t i = 0; i < size; i++) {
	const ValueType d = p1[i] - m_points[idx_p2][i];
	s += d*d;
      }
      return s;
    }

    // Returns the dim'th component of the idx'th point in the class:
    inline ValueType kdtree_get_pt(const SizeType idx, size_t dim) const 
    {
      return m_points[idx][dim];
    }

    // Optional bounding-box computation: return false to default to a standard bbox computation loop.
    //   Return true if the BBOX was already computed by the class and returned in "bb" so it can be avoided to redo it again.
    //   Look at bb.size() to find out the expected dimensionality (e.g. 2 or 3 for point clouds)
    template <class BBOX>
    bool kdtree_get_bbox(BBOX &bb) const 
    {
      return false;
    }

    /** @} */
    
    
    /**
    * strategy and nrOfPoints are ignored
    *  == evalAtPoint_simple
    **/
    template <class DOFVector>
    Value_t<DOFVector> evalAtPoint(DOFVector const& vec, PointType const& x)
    {
      Value_t<DOFVector> value = 0;
      
      auto const& grid = vec.getFeSpace().gridView().grid();
      DataType elementSeed;
      if (getNearestElement(x, elementSeed)) {
        auto element = grid.entity(elementSeed);
        auto lambda = element.geometry().local(x);
        value = localEval(vec, element, lambda);
      }
      return value;
    }
    
  private:    
    /// evaluate DOFVector at barycentric coordinates \p lambda in element
    template <class FE, class Value, class Element, class LocalCoords>
    Value localEval(DOFVector<FE, Value> const& vec,
                    Element const& element,
                    LocalCoords const& lambda)
    {
      localView.bind(element);
      localIndexSet.bind(localView);
      
      auto const& localFiniteElem = localView.tree().finiteElement();
      auto const& localBasis = localFiniteElem.localBasis();
      const size_t nBasisFct = localFiniteElem.size();

      std::vector<Dune::FieldVector<double,1> > shapeValues;
      localBasis.evaluateFunction(lambda, shapeValues);
      
      Value data = 0;
      for (size_t j = 0; j < nBasisFct; ++j) {
        auto const global_idx = localIndexSet.index(j);
        data += vector[global_idx] * shapeValues[j];
      }
      
      return data;
    }
    
    
    bool getNearestElement(PointType const& x, DataType& elementSeed)
    {
      const size_t nrOfPoints = 1;
      std::vector<DegreeOfFreedom> ret_indexes(nrOfPoints);
      std::vector<double> out_dists_sqr(nrOfPoints);
      nanoflann::KNNResultSet<double, DegreeOfFreedom> resultSet(nrOfPoints);
      
      resultSet.init(&ret_indexes[0], &out_dists_sqr[0]);
      index->findNeighbors(resultSet, x.begin(), nanoflann::SearchParams());
      
      elementSeed = m_data[ret_indexes[0]];      
      return true;
    }
    
  private:
    typename FeSpace::LocalView     localView;
    typename FeSpace::LocalIndexSet localIndexSet;

  }; // end of KDTreeMeshAdaptor

  

  template <class FeSpace>
  auto provideKDTree(FeSpace const& feSpace)
  {
    using KD_Tree = KDTreeMeshAdaptor<FeSpace>;

    return make_unique<KD_Tree>(feSpace);
  }
  
} // end namespace AMDiS
