#include <iostream>
#include <ctime>
#include <cmath>

#include <dune/amdis/AMDiS.hpp>
#include <dune/amdis/AdaptInstationary.hpp>
#include <dune/amdis/ProblemStat.hpp>
#include <dune/amdis/ProblemInstat.hpp>
#include <dune/amdis/Terms.hpp>

#ifdef DOW
#undef DOW
#endif
#define DOW AMDIS_DOW

#ifndef STRATEGY
#define STRATEGY 1
#endif

using namespace AMDiS;

using Mesh = Dune::YaspGrid<AMDIS_DIM>;
// using Mesh = Dune::AlbertaGrid<AMDIS_DIM, AMDIS_DOW>;

// 3 components: velocity with polynomial degree 2 and pressure with polynomial degree 1
using StokesParam   = DefaultTraitsMesh<Mesh, AMDIS_DIM, AMDIS_DOW, 2, 2, 1>;
using StokesProblem = ProblemStat<StokesParam>;
using StokesProblemInstat = ProblemInstat<StokesParam>;

int main(int argc, char** argv)
{
    AMDiS::init(argc, argv);

    StokesProblem prob("stokes");
    prob.initialize(INIT_ALL);

    StokesProblemInstat probInstat("stokes", prob);
    probInstat.initialize(INIT_UH_OLD);

    double viscosity = 1.0;
    double density = 1.0;
    double vel = 1.0;
    Parameters::get("stokes->viscosity", viscosity);
    Parameters::get("stokes->density", density);
    Parameters::get("stokes->boundary velocity", vel);

    AdaptInfo adaptInfo("adapt", prob.getNumComponents());

    // define the differential operators
    using Op = StokesProblem::OperatorType;
    For<0,DOW>::loop([&](auto const _i)
    {
	// <1/tau * u_i, v_i>
        auto opTime = Op::zot( density );
        auto opTimeOld = Op::zot( valueOf(prob.getSolution(_i), density) );
        prob.addMatrixOperator(opTime, _i, _i, probInstat.getInvTau());
        prob.addVectorOperator(opTimeOld,  _i, probInstat.getInvTau());

#if STRATEGY == 1
        // <(u * nabla)u_i^old, v_i>
        For<0, DOW>::loop([&](auto const _j)
        {
          auto opNonlin = Op::zot( derivativeOf(prob.getSolution(_i), _j, density) );
          prob.addMatrixOperator(opNonlin, _i, _j);
        });
#elif STRATEGY == 2
        // <(u^old * nabla)u_i, v_i>
        For<0, DOW>::loop([&](auto const _j)
        {
          auto opNonlin = Op::fot( valueOf(prob.getSolution(_j), density), _j, GRD_PHI );
          prob.addMatrixOperator(opNonlin, _i, _i);
        });
#else
        // <(u^old * nabla)u_i^old, v_i>
        For<0, DOW>::loop([&](auto const _j)
        {
          auto opNonlin = Op::zot( density * valueOf(prob.getSolution(_j)) * derivativeOf(prob.getSolution(_i), _j) );
          prob.addVectorOperator(opNonlin, _i);
        });
#endif

	auto opL = std::make_pair( Op::sot( viscosity ), false ); // <viscosity*grad(u_i), grad(v_i)>
	auto opP = std::make_pair( Op::fot( 1.0, _i, GRD_PSI ), false); // <p, d_i(v_i)>
	auto opDiv = std::make_pair( Op::fot( 1.0, _i, GRD_PHI ), false); // <d_i(u_i), q>

        const int i = _i;
        prob.addMatrixOperator({{
          {{i,  i}, opL}, {{i,DOW}, opP},
          {{DOW,i}, opDiv}
        }});
    });

    auto opZero = std::make_pair( Op::zot( 0.0 ), false );
    prob.addMatrixOperator({{ {{DOW,  DOW}, opZero} }});


    // define boundary regions
    auto left = [](auto const& x) { return x[0] < 1.e-8; };
    auto box  = [](auto const& x) { return x[0] > 1.0 - 1.e-8 || x[1] < 1.e-8 || x[1] > 1.0 - 1.e-8; };
    auto corner = [](auto const& x) { return x[0] == 0.0 && x[1] == 0.0; };

    // define boundary values
    auto zero      = [](auto const& x) { return 0.0; };
    auto parabolic = [vel](auto const& x) { return vel*4.0*x[1]*(1.0 - x[1]); };

    // set boundary conditions
    for (size_t i = 0; i < DOW; ++i)
	prob.addDirichletBC(box, i, i, zero);

    prob.addDirichletBC(left, 0, 0, zero);
    prob.addDirichletBC(left, 1, 1, parabolic);

    // pressure fixation
    prob.addDirichletBC(corner, DOW, DOW, zero);

    *prob.getSolution() = 0.0;

    AdaptInstationary adapt("adapt", prob, adaptInfo, probInstat, adaptInfo);
    adapt.adapt();

    // output solution
    prob.writeFiles(adaptInfo);

    AMDiS::finalize();
    return 0;
}
