#include <iostream>
#include <ctime>
#include <cmath>

#include <dune/amdis/AMDiS.hpp>
#include <dune/amdis/ProblemStat.hpp>
#include <dune/amdis/Terms.hpp>

#ifdef DOW
#undef DOW
#endif
#define DOW AMDIS_DOW

using namespace AMDiS;

// 3 components: velocity with polynomial degree 2 and pressure with polynomial degree 1
using StokesParam   = ProblemStatTraits<AMDIS_DIM, AMDIS_DOW, 2, 2, 1>;
using StokesProblem = ProblemStat<StokesParam>;

int main(int argc, char** argv)
{
    AMDiS::init(argc, argv);

    StokesProblem prob("stokes");
    prob.initialize(INIT_ALL);

    double viscosity = 1.0;
    Parameters::get("stokes->viscosity", viscosity);

    // define the differential operators
    using Op = StokesProblem::OperatorType;
    for (int i = 0; i < DOW; ++i) {
        // <viscosity*grad(u_i), grad(v_i)>
        Op* opL = new Op;
        opL->addSOT( constant(viscosity) );
        prob.addMatrixOperator(*opL, i, i);

        // <p, d_i(v_i)>
        Op* opB = new Op;
        opB->addFOT( constant(1.0), i, GRD_PSI );
        prob.addMatrixOperator(*opB, i, DOW);

        // <d_i(u_i), q>
        Op* opDiv = new Op;
        opDiv->addFOT( constant(1.0), i, GRD_PHI );
        prob.addMatrixOperator(*opDiv, DOW, i);
    }

    Op* opZero = new Op;
    opZero->addZOT( constant(0.0) );
    prob.addMatrixOperator(*opZero, DOW, DOW);


    // define boundary regions
    auto left = [](auto const& x) { return x[0] < 1.e-8; };
    auto box  = [](auto const& x) { return x[0] > 1.0 - 1.e-8 || x[1] < 1.e-8 || x[1] > 1.0 - 1.e-8; };

    // define boundary values
    auto zero      = [](auto const& x) { return 0.0; };
    auto parabolic = [](auto const& x) { return x[1]*(1.0 - x[1]); };

    // set boundary conditions
    for (size_t i = 0; i < DOW; ++i)
	    prob.addDirichletBC(box, i, i, zero);

    prob.addDirichletBC(left, 0, 0, zero);
    prob.addDirichletBC(left, 1, 1, parabolic);

    prob.addDirichletBC([](auto const& x) { return x[0] < 1.e-8 && x[1] < 1.e-8; }, 2, 2, zero);

    // set initial values for the solver (maybe not necessary)
    *prob.getSolution() = 0.0;

    AdaptInfo adaptInfo("adapt", prob.getNumComponents());

    // assemble and solve system
    prob.buildAfterCoarsen(adaptInfo, Flag(0));
    prob.solve(adaptInfo);

    // output solution
    prob.writeFiles(adaptInfo);

    AMDiS::finalize();
    return 0;
}
